# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BLEU metric implementation."""

import collections
import math

import tensorflow as tf
from tensorflow import keras

from keras_nlp.src.api_export import keras_nlp_export
from keras_nlp.src.utils.tf_utils import tensor_to_list

REPLACE_SUBSTRINGS = [
    ("<skipped>", ""),
    ("-\n", ""),
    ("\n", " "),
    ("&quot;", '"'),
    ("&amp;", "&"),
    ("&lt;", "<"),
    ("&gt;", ">"),
]


REGEX_PATTERNS = [
    # language-dependent part (assuming Western languages)
    (r"([\{-\~\[-\` -\&\(-\+\:-\@\/])", r" \1 "),
    # tokenize period and comma unless preceded by a digit
    (r"([^0-9])([\.,])", r"\1 \2 "),
    # tokenize period and comma unless followed by a digit
    (r"([\.,])([^0-9])", r" \1 \2"),
    # tokenize dash when preceded by a digit
    (r"([0-9])(-)", r"\1 \2 "),
    # If last character is "." or ",", add space.
    (r"[\.,]$", r" \0 \1"),
    # one space only between words
    (r"\s+", r" "),
]


@keras_nlp_export("keras_nlp.metrics.Bleu")
class Bleu(keras.metrics.Metric):
    """BLEU metric.

    This class implements the BLEU metric. BLEU is generally used to evaluate
    machine translation systems. By default, this implementation replicates
    SacreBLEU, but user-defined tokenizers can be passed to deal with other
    languages.

    For BLEU score, we count the number of matching n-grams in the candidate
    translation and the reference text. We find the "clipped count" of matching
    n-grams so as to not give a high score to a (reference, prediction) pair
    with redundant, repeated tokens. Secondly, BLEU score tends to reward
    shorter predictions more, which is why a brevity penalty is applied to
    penalise short predictions. For more details, see the following article:
    https://cloud.google.com/translate/automl/docs/evaluate#bleu.

    Note on input shapes:
    For unbatched inputs, `y_pred` should be a tensor of shape `()`, and
    `y_true` should be a tensor of shape `(num_references,)`. For batched
    inputs, `y_pred` should be a tensor of shape `(batch_size,)`,
    and `y_true` should be a tensor of shape `(batch_size, num_references)`. In
    case of batched inputs, `y_true` can also be a ragged tensor of shape
    `(batch_size, None)` if different samples have different number of
    references.

    Args:
        tokenizer: callable. A function that takes a string `tf.RaggedTensor`
            (of any shape), and tokenizes the strings in the tensor. If the
            tokenizer is not specified, the default tokenizer is used. The
            default tokenizer replicates the behaviour of SacreBLEU's
            `"tokenizer_13a"` tokenizer
            (https://github.com/mjpost/sacrebleu/blob/v2.1.0/sacrebleu/tokenizers/tokenizer_13a.py).
        max_order: int. The maximum n-gram order to use. For example, if
            `max_order` is set to 3, unigrams, bigrams, and trigrams will be
            considered. Defaults to 4.
        smooth: bool. Whether to apply Lin et al. 2004 smoothing to the BLEU
            score. Adds 1 to the matched n-gram count (i.e., numerator) and 1
            to the total n-gram count (i.e., denominator) for every order while
            calculating precision. Defaults to False.
        dtype: string or tf.dtypes.Dtype. Precision of metric computation. If
               not specified, it defaults to tf.float32.
        name: string. Name of the metric instance.
        **kwargs: Other keyword arguments.

    References:
        - [Papineni et al., 2002](https://aclanthology.org/P02-1040/)
        - [SacreBLEU](https://github.com/mjpost/sacrebleu)
        - [Lin et al., 2004](https://aclanthology.org/P04-1077/)
    """

    def __init__(
        self,
        tokenizer=None,
        max_order=4,
        smooth=False,
        dtype=None,
        name="bleu",
        **kwargs,
    ):
        super().__init__(name=name, dtype=dtype, **kwargs)

        if not tf.as_dtype(self.dtype).is_floating:
            raise ValueError(
                "`dtype` must be a floating point type. "
                f"Received: dtype={dtype}"
            )

        self.tokenizer = tokenizer
        self.max_order = max_order
        self.smooth = smooth

        self._matches = self.add_weight(
            shape=(self.max_order,),
            name="bleu_matches",
            initializer="zeros",
            dtype=self.dtype,
        )
        self._possible_matches = self.add_weight(
            shape=(self.max_order,),
            name="bleu_possible_matches",
            initializer="zeros",
            dtype=self.dtype,
        )
        self._translation_length = self.add_weight(
            name="bleu_translation_length",
            initializer="zeros",
            dtype=self.dtype,
        )
        self._reference_length = self.add_weight(
            name="bleu_reference_length",
            initializer="zeros",
            dtype=self.dtype,
        )
        self._bleu = self.add_weight(
            name="bleu",
            initializer="zeros",
            dtype=self.dtype,
        )

    def _tokenizer(self, inputs):
        """
        Tokenizes the input strings. By default, replicates the behaviour of
        SacreBLEU's default tokenizer, namely, `tokenizer_13a`.
        """
        if self.tokenizer:
            return self.tokenizer(inputs)

        for pattern, replacement in REPLACE_SUBSTRINGS + REGEX_PATTERNS:
            inputs = tf.strings.regex_replace(
                input=inputs,
                pattern=pattern,
                rewrite=replacement,
                replace_global=True,
                name=None,
            )
        inputs = tf.strings.split(inputs)
        return inputs

    def _get_ngrams(self, segment, max_order):
        """Extracts all n-grams up to a given maximum order from an input segment.

        Uses Python ops. Inspired from
        https://github.com/tensorflow/nmt/blob/master/nmt/scripts/bleu.py.

        Args:
            segment: list. Text segment from which n-grams will be
                extracted.
            max_order: int. Maximum length in tokens of the n-grams returned
                by this method.
        """
        ngram_counts = collections.Counter()
        for order in range(1, max_order + 1):
            for i in range(0, len(segment) - order + 1):
                ngram = tuple(segment[i : i + order])
                ngram_counts[ngram] += 1
        return ngram_counts

    def _corpus_bleu(
        self,
        reference_corpus,
        translation_corpus,
        matches_by_order,
        possible_matches_by_order,
        translation_length,
        reference_length,
        max_order=4,
        smooth=False,
    ):
        """Corpus BLEU implementation using Python ops.

        Computes BLEU score of translated segments against one or more
        references. Inspired from
        https://github.com/tensorflow/nmt/blob/master/nmt/scripts/bleu.py.

        Args:
            reference_corpus: list of lists of references for each
                translation. Each reference should be tokenized into a list
                of tokens.
            translation_corpus: list of translations to score. Each
                translation should be tokenized into a list of tokens.
            matches_by_order: list of floats containing the initial number
                of matches for each order.
            possible_matches_by_order: list of floats containing the initial
                number of possible matches for each order.
            translation_length: float. Initial number of tokens in all the
                translations.
            reference_length: float. Initial number of tokens in all the
                references.
            max_order: int. Maximum n-gram order to use when computing
                BLEU score.
            smooth: boolean. Whether or not to apply Lin et al. 2004
                smoothing.
        """
        for references, translation in zip(
            reference_corpus, translation_corpus
        ):
            reference_length += min(len(r) for r in references)
            translation_length += len(translation)

            merged_ref_ngram_counts = collections.Counter()
            for reference in references:
                merged_ref_ngram_counts |= self._get_ngrams(
                    reference, max_order
                )
            translation_ngram_counts = self._get_ngrams(translation, max_order)
            overlap = translation_ngram_counts & merged_ref_ngram_counts
            for ngram in overlap:
                matches_by_order[len(ngram) - 1] += overlap[ngram]
            for order in range(1, max_order + 1):
                possible_matches = len(translation) - order + 1
                if possible_matches > 0:
                    possible_matches_by_order[order - 1] += possible_matches

        precisions = [0] * max_order
        for i in range(0, max_order):
            if smooth:
                precisions[i] = (matches_by_order[i] + 1.0) / (
                    possible_matches_by_order[i] + 1.0
                )
            else:
                if possible_matches_by_order[i] > 0:
                    precisions[i] = (
                        float(matches_by_order[i])
                        / possible_matches_by_order[i]
                    )
                else:
                    precisions[i] = 0.0

        if min(precisions) > 0:
            p_log_sum = sum((1.0 / max_order) * math.log(p) for p in precisions)
            geo_mean = math.exp(p_log_sum)
        else:
            geo_mean = 0

        ratio = float(translation_length) / reference_length

        if ratio > 1.0:
            bp = 1.0
        else:
            bp = math.exp(1 - 1.0 / ratio)

        bleu = geo_mean * bp

        return (
            bleu,
            matches_by_order,
            possible_matches_by_order,
            translation_length,
            reference_length,
        )

    def _calculate_bleu_score(self, references, translation):
        references = tensor_to_list(references)
        translation = tensor_to_list(translation)

        matches = self._matches.numpy()
        possible_matches = self._possible_matches.numpy()
        translation_length = self._translation_length.numpy()
        reference_length = self._reference_length.numpy()

        (
            bleu_score,
            matches,
            possible_matches,
            translation_length,
            reference_length,
        ) = self._corpus_bleu(
            reference_corpus=references,
            translation_corpus=translation,
            matches_by_order=matches,
            possible_matches_by_order=possible_matches,
            translation_length=translation_length,
            reference_length=reference_length,
            max_order=self.max_order,
            smooth=self.smooth,
        )
        return (
            tf.constant(bleu_score, dtype=self.dtype),
            tf.constant(matches, dtype=self.dtype),
            tf.constant(possible_matches, dtype=self.dtype),
            tf.constant(translation_length, dtype=self.dtype),
            tf.constant(reference_length, dtype=self.dtype),
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        def validate_and_fix_rank(inputs, tensor_name, base_rank=0):
            if not isinstance(inputs, (tf.Tensor, tf.RaggedTensor)):
                inputs = tf.convert_to_tensor(inputs)

            if inputs.shape.rank == base_rank:
                return inputs[tf.newaxis]
            elif inputs.shape.rank == base_rank + 1:
                return inputs
            elif inputs.shape.rank == base_rank + 2:
                if tf.shape(inputs)[-1] != 1:
                    raise ValueError(
                        f"{tensor_name} is of rank {input.shape.rank}. The "
                        f"last dimension must be of size 1."
                    )
                return tf.squeeze(inputs, axis=-1)
            else:
                raise ValueError(
                    f"{tensor_name} must be of rank {base_rank}, {base_rank+1} "
                    f"or {base_rank+2}. Found rank: {inputs.shape.rank}"
                )

        y_true = validate_and_fix_rank(y_true, "y_true", 1)
        y_pred = validate_and_fix_rank(y_pred, "y_pred", 0)

        # Tokenize the inputs.
        y_true = self._tokenizer(y_true)
        y_pred = self._tokenizer(y_pred)

        (
            bleu_score,
            matches,
            possible_matches,
            translation_length,
            reference_length,
        ) = tf.py_function(
            func=self._calculate_bleu_score,
            inp=[y_true, y_pred],
            Tout=[self.dtype, self.dtype, self.dtype, self.dtype, self.dtype],
        )

        self._matches.assign(matches)
        self._possible_matches.assign(possible_matches)
        self._translation_length.assign(translation_length)
        self._reference_length.assign(reference_length)
        self._bleu.assign(bleu_score)

    def result(self):
        return self._bleu

    def reset_state(self):
        self._matches.assign(
            tf.zeros(shape=(self.max_order,), dtype=self.dtype)
        )
        self._possible_matches.assign(
            tf.zeros(shape=(self.max_order,), dtype=self.dtype)
        )
        self._translation_length.assign(0.0)
        self._reference_length.assign(0.0)
        self._bleu.assign(0.0)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "tokenizer": self.tokenizer,
                "max_order": self.max_order,
                "smooth": self.smooth,
            }
        )
        return config




# Copyright 2022 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OPT backbone model."""

import copy

import tensorflow as tf
from tensorflow import keras
from tensorflow.experimental import dtensor
from tensorflow.experimental.dtensor import Layout

from keras_nlp.src.api_export import keras_nlp_export
from keras_nlp.src.layers.token_and_position_embedding import (
    TokenAndPositionEmbedding,
)
from keras_nlp.src.layers.transformer_decoder import TransformerDecoder
from keras_nlp.src.models.backbone import Backbone
from keras_nlp.src.models.opt.opt_presets import backbone_presets
from keras_nlp.src.utils.python_utils import classproperty


def opt_kernel_initializer(stddev=0.02):
    return keras.initializers.TruncatedNormal(stddev=stddev)


@keras_nlp_export("keras_nlp.models.OPTBackbone")
class OPTBackbone(Backbone):
    """An OPT decoder network.

    This class implements a Transformer-based decoder model as described in
    ["OPT: Open Pre-trained Transformer Language Models"](https://arxiv.org/abs/2205.01068).
    The default constructor gives a fully customizable, randomly initialized OPT
    model with any number of layers, heads, and embedding dimensions. To load
    preset architectures and weights, use the `from_preset()` constructor.

    Disclaimer: Pre-trained models are provided on an "as is" basis, without
    warranties or conditions of any kind. The underlying model is provided by a
    third party and subject to a separate license, available
    [here](https://github.com/facebookresearch/fairseq/).

    Args:
        vocabulary_size: int. The size of the token vocabulary.
        num_layers: int. The number of transformer decoder layers.
        num_heads: int. The number of attention heads for each transformer.
            The hidden size must be divisible by the number of attention heads.
        hidden_dim: int. The hidden size of the transformer decoder layers.
        intermediate_dim: int. The output dimension of the first Dense layer in
            a two-layer feedforward network for each transformer decoder layer.
        dropout: float. Dropout probability for the Transformer decoder.
        max_sequence_length: int. The maximum sequence length that this decoder
            can consume. If `None`, `max_sequence_length` uses the value from
            sequence length. This determines the variable shape for positional
            embeddings.

    Examples:
    ```python
    input_data = {
        "token_ids": tf.ones(shape=(1, 12), dtype=tf.int64),
        "padding_mask": tf.constant(
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0], shape=(1, 12)
        ),
    }

    # Pretrained OPT decoder
    model = keras_nlp.models.OPTBackbone.from_preset("opt_125m_en")
    model(input_data)

    # Randomly initialized OPT decoder model with a custom config
    model = keras_nlp.models.OPTBackbone(
        vocabulary_size=50265,
        num_layers=4,
        num_heads=4,
        hidden_dim=256,
        intermediate_dim=512,
        max_sequence_length=128,
    )
    model(input_data)
    ```
    """

    def __init__(
        self,
        vocabulary_size,
        num_layers,
        num_heads,
        hidden_dim,
        intermediate_dim,
        dropout=0.1,
        max_sequence_length=2048,
        **kwargs,
    ):
        # Decoder inputs.
        token_ids = keras.Input(shape=(None,), dtype="int32", name="token_ids")
        padding_mask = keras.Input(
            shape=(None,), dtype="int32", name="padding_mask"
        )

        # Embed tokens and positions.
        x = TokenAndPositionEmbedding(
            vocabulary_size=vocabulary_size,
            sequence_length=max_sequence_length,
            embedding_dim=hidden_dim,
            embeddings_initializer=opt_kernel_initializer(),
            name="embeddings",
        )(token_ids)

        # Apply successive transformer decoder blocks.
        for i in range(num_layers):
            x = TransformerDecoder(
                intermediate_dim=intermediate_dim,
                num_heads=num_heads,
                dropout=dropout,
                activation="relu",
                layer_norm_epsilon=1e-5,
                normalize_first=True,
                kernel_initializer=opt_kernel_initializer(),
                name=f"transformer_layer_{i}",
            )(x, decoder_padding_mask=padding_mask)

        # Add a final layer norm.
        x = keras.layers.LayerNormalization(
            name="layer_norm",
            axis=-1,
            epsilon=1e-5,
            dtype=tf.float32,
        )(x)

        # Instantiate using Functional API Model constructor
        super().__init__(
            inputs={
                "token_ids": token_ids,
                "padding_mask": padding_mask,
            },
            outputs=x,
            **kwargs,
        )

        # All references to `self` below this line
        self.vocabulary_size = vocabulary_size
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.dropout = dropout
        self.max_sequence_length = max_sequence_length

    def get_config(self):
        return {
            "vocabulary_size": self.vocabulary_size,
            "num_layers": self.num_layers,
            "num_heads": self.num_heads,
            "hidden_dim": self.hidden_dim,
            "intermediate_dim": self.intermediate_dim,
            "dropout": self.dropout,
            "max_sequence_length": self.max_sequence_length,
        }

    @property
    def token_embedding(self):
        return self.get_layer("embeddings").token_embedding

    @classproperty
    def presets(cls):
        return copy.deepcopy(backbone_presets)

    @classmethod
    def create_layout_map(cls, mesh):
        """Create a DTensor layout map for an OPTBackbone.

        Given a DTensor mesh describing a list of devices, this method returns a
        DTensor layout map for creating a `keras_nlp.models.OPTBackbone`
        instance. This mapping describes how to distribute all model weights
        across multiple devices. For an overview of DTensor concepts, see
        [this guide](https://www.tensorflow.org/guide/dtensor_overview).

        Args:
            mesh: A 2D `tf.experimental.dtensor.Mesh` describing the arrangement
                of devices for running distributed computation. The
                first dimension in the mesh is expected to be for data parallel
                distribution, and the second for model parallel distribution.

        Returns:
            A `tf.keras.dtensor.experimental.LayoutMap` which contains the
            proper layout to weights mapping for the model parallel setting.

        Examples:
        ```python
        keras.backend.experimental.enable_tf_random_generator()
        keras.utils.set_random_seed(1337)

        # Update both dimensions below for a multi-device setting.
        mesh = dtensor.create_mesh([("batch", 1), ("model", 1)])
        layout_map = keras_nlp.models.OPTBackbone.create_layout_map(mesh)

        with layout_map.scope():
            model = keras_nlp.models.OPTBackbone.from_preset("opt_125m_en")
        ```
        """
        # We assert the mesh is 2D, and assume the first mesh dim is for data
        # parallel and the second dim is for model parallel.
        mesh_shape = mesh.shape()
        if len(mesh_shape) != 2:
            raise ValueError(
                f"Expect to create layout based on 2D mesh, received {mesh}"
            )
        _, model_dim = mesh.dim_names
        unshard_dim = dtensor.UNSHARDED

        layout_map = keras.dtensor.experimental.LayoutMap(mesh=mesh)
        # Embedding sharding
        layout_map[r".*embeddings"] = Layout([unshard_dim, model_dim], mesh)

        # Transformer block sharding
        layout_map[r".*_(query|key|value)_dense.kernel"] = Layout(
            [unshard_dim, unshard_dim, model_dim], mesh
        )
        layout_map[r".*_(query|key|value)_dense.bias"] = Layout(
            [model_dim, unshard_dim], mesh
        )
        layout_map[r".*_feedforward_intermediate_dense.kernel"] = Layout(
            [unshard_dim, model_dim], mesh
        )
        layout_map[r".*_feedforward_intermediate_dense.bias"] = Layout(
            [model_dim], mesh
        )
        layout_map[r".*_feedforward_output_dense.kernel"] = Layout(
            [model_dim, unshard_dim], mesh
        )
        layout_map[r".*_feedforward_output_dense.bias"] = Layout(
            [unshard_dim], mesh
        )
        return layout_map




# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Beam Sampler."""

import tensorflow as tf
from tensorflow import keras
from tensorflow.compiler.tf2xla.python.xla import dynamic_update_slice

from keras_nlp.src.api_export import keras_nlp_export
from keras_nlp.src.samplers.sampler import Sampler
from keras_nlp.src.samplers.sampler import call_args_docstring
from keras_nlp.src.utils.python_utils import format_docstring


@format_docstring(call_args=call_args_docstring)
@keras_nlp_export("keras_nlp.samplers.BeamSampler")
class BeamSampler(Sampler):
    """Beam Sampler class.

    This sampler implements beam search algorithm. At each time-step, beam
    search keeps the beams (sequences) of the top `num_beams` highest
    accumulated probabilities, and uses each one of the beams to predict
    candidate next tokens.

    Args:
        num_beams: int. The number of beams that should be kept at each
            time-step. `num_beams` should be strictly positive.
        return_all_beams: bool. When set to `True`, the sampler will return all
            beams and their respective probabilities score.

    Call arguments:
        {{call_args}}

    Examples:
    Return only the beam with the highest accumulated probability.
    ```python
    # Use a simple alphabet of lowercase characters with ids in range [0, 25].
    int_lookup = {i: chr(i + ord('a')) for i in range(26)}
    char_lookup = {v: k for k, v in int_lookup.items()}
    batch_size, length, vocab_size = 1, 12, len(int_lookup)

    def next(prompt, cache, index):
        prompt_batch_size = tf.shape(prompt)[0]
        hidden_states = tf.ones((prompt_batch_size, 10))
        # A uniform distribution over our alphabet.
        logits = tf.ones((prompt_batch_size, vocab_size))
        return logits, hidden_states, cache

    output = keras_nlp.samplers.BeamSampler()(
        next=next,
        prompt=tf.fill((batch_size, length), char_lookup["z"]),
        index=5,
    )
    print(["".join([int_lookup[i] for i in s]) for s in output.numpy()])
    # >>> ['zzzzzeeeeeee']
    ```

    Return all beams and their probabilities.
    ```python
    # Use a simple alphabet of lowercase characters with ids in range [0, 25].
    int_lookup = {i: chr(i + ord('a')) for i in range(26)}
    char_lookup = {v: k for k, v in int_lookup.items()}
    batch_size, length, vocab_size = 1, 8, len(int_lookup)

    def next(prompt, cache, index):
        prompt_batch_size = tf.shape(prompt)[0]
        hidden_states = tf.ones((prompt_batch_size, 10))
        # A uniform distribution over our alphabet.
        logits = tf.ones((batch_size, vocab_size))
        return logits, hidden_states, cache

    beams, probs = keras_nlp.samplers.BeamSampler(return_all_beams=True)(
        next=next,
        prompt=tf.fill((batch_size, length,), char_lookup['z']),
        index=5,
    )

    print(beams.shape)
    # >>> (1, 5, 8)
    print(probs.shape)
    # >>> (1, 5)
    print(["".join([int_lookup[i] for i in s]) for s in beams[0].numpy()])
    # >>> ['zzzzzeee', 'zzzzzeed', 'zzzzzeec', 'zzzzzeea', 'zzzzzeeb']
    ```
    """

    def __init__(
        self,
        num_beams=5,
        return_all_beams=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_beams = num_beams
        self.return_all_beams = return_all_beams

    def __call__(
        self,
        next,
        prompt,
        cache=None,
        index=0,
        mask=None,
        end_token_id=None,
        hidden_states=None,
    ):
        batch_size, max_length = tf.shape(prompt)[0], tf.shape(prompt)[1]
        # Make sure max length and start index are the same dtype.
        index = tf.cast(index, max_length.dtype)

        def create_beams(x):
            """Add initial beam state."""
            return tf.repeat(x, self.num_beams, axis=0)

        def flatten_beams(x):
            """Combine the beam dim and batch dim."""
            flat_shape = [batch_size * self.num_beams] + x.shape.as_list()[2:]
            return tf.reshape(x, shape=flat_shape)

        def unflatten_beams(x):
            """Separate the beam dim and batch dim."""
            unflat_shape = [batch_size, self.num_beams] + x.shape.as_list()[1:]
            return tf.reshape(x, shape=unflat_shape)

        if mask is None:
            mask = tf.zeros_like(prompt, dtype=tf.bool)
        else:
            mask = tf.cast(mask, dtype=tf.bool)
        # `tf.while_loop` will not accept `None` as a value for `loop_vars`.
        cache = () if cache is None else cache
        # Add extra sequences for each beam.
        prompt, mask = create_beams(prompt), create_beams(mask)
        cache = tf.nest.map_structure(create_beams, cache)
        # Setup the initial beam log-likelihoods.
        # On the first loop, make sure only the original beam is considered.
        log_probs = tf.constant([[0.0] + [-1e9] * (self.num_beams - 1)])
        log_probs = flatten_beams(tf.repeat(log_probs, batch_size, axis=0))

        def cond(prompt, cache, index, log_probs):
            if end_token_id is None:
                return True
            # Stop if all sequences have produced a *new* end_token_id.
            end_tokens = (prompt == end_token_id) & (~mask)
            prompt_done = tf.reduce_any(end_tokens, axis=-1)
            return not tf.reduce_all(prompt_done)

        def body(prompt, cache, index, log_probs):
            # Compute the softmax distribution for the next token.
            logits, _, cache = next(prompt, cache, index)
            vocab_size = tf.shape(logits)[-1]
            probs = keras.activations.softmax(logits / self.temperature)

            # Compute the running log-likelihood of each new candidate.
            next_log_probs = tf.math.log(probs) + log_probs[..., tf.newaxis]
            # Reshape `preds` to shape `(batch_size, num_beams * vocab_size)`.
            next_log_probs = tf.reshape(next_log_probs, shape=[batch_size, -1])

            # Compute the top beam indices and next tokens.
            next_log_probs, indices = tf.math.top_k(
                next_log_probs, k=self.num_beams, sorted=False
            )
            beam_indices = indices // vocab_size
            next_token = flatten_beams(indices % vocab_size)
            # Ensure shape is `[None]`, otherwise it causes issues after
            # converting to TFLite.
            next_token = tf.ensure_shape(next_token, [None])
            # We need `ensure_shape` as `top_k` will change the static shape.
            next_log_probs = flatten_beams(next_log_probs)
            log_probs = tf.ensure_shape(next_log_probs, log_probs.shape)

            def gather_beams(x):
                x = unflatten_beams(x)
                x = tf.gather(x, beam_indices, axis=1, batch_dims=1)
                return flatten_beams(x)

            prompt = gather_beams(prompt)
            cache = tf.nest.map_structure(gather_beams, cache)

            # Update each beam with the next token.
            next_token = tf.cast(next_token, prompt.dtype)
            # Don't overwrite anywhere mask is True.
            next_token = tf.where(mask[:, index], prompt[:, index], next_token)
            # Update the prompt with the next token.
            next_token = next_token[:, tf.newaxis]
            prompt = dynamic_update_slice(prompt, next_token, [0, index])
            # Return the iteration of the loop state.
            return (prompt, cache, index + 1, log_probs)

        prompt, _, _, log_probs = tf.while_loop(
            cond=cond,
            body=body,
            loop_vars=(prompt, cache, index, log_probs),
            maximum_iterations=(max_length - index),
        )

        all_prompts = unflatten_beams(prompt)
        all_log_probs = unflatten_beams(log_probs)

        if self.return_all_beams:
            sorted_indices = tf.argsort(
                all_log_probs, axis=-1, direction="DESCENDING"
            )
            sorted_log_probs = tf.gather(
                all_log_probs, sorted_indices, axis=-1, batch_dims=1
            )
            sorted_prompts = tf.gather(
                all_prompts, sorted_indices, axis=1, batch_dims=1
            )
            return sorted_prompts, sorted_log_probs
        else:
            # Gather the top beam at each batch index.
            top_beams = tf.math.argmax(all_log_probs, axis=-1)[:, tf.newaxis]
            prompt = tf.gather(all_prompts, top_beams, axis=1, batch_dims=1)
            return tf.squeeze(prompt, axis=1)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_beams": self.num_beams,
                "return_all_beams": self.return_all_beams,
            }
        )
        return config

